module MediServ::API
  require 'dry-types'
  require 'dry-struct'

  require_relative "format"

  module Types
    include Dry.Types()
  end

  class Struct < Dry::Struct
    include Format
  end

  class Recipient < Struct
    # Nummer des Empfängers
    attribute :nummer, Types::Strict::Integer
    # Art des Empfängers (Garant oder Gesetz_Vertreter)
    attribute :art, Types::Strict::String
    # Anrede in der Sprache der Rechnung
    attribute :anrede, Types::Strict::String
    # Titel in der Sprache der Rechnung
    attribute :titel, Types::Strict::String
    # Name
    attribute :name, Types::Strict::String
    # Vorname
    attribute :vorname, Types::Strict::String
    # Strasse mit Nummer
    attribute :strasse, Types::Strict::String.optional.default(nil)
    # Land (CH / D / F / ...)
    attribute :land, Types::Strict::String
    # Postleitzahl
    attribute :plz, Types::Strict::String
    # Ort
    attribute :ort, Types::Strict::String
    # Geburtsdatum
    attribute :geburtsdatum, Types::Strict::Date
  end

  class Patient < Struct
    # Patientennummer
    attribute :nummer, Types::Strict::Integer
    # Anrede in der Sprache der Rechnung
    attribute :anrede, Types::Strict::String
    # Titel in der Sprache der Rechnung
    attribute :titel, Types::Strict::String
    # Name
    attribute :name, Types::Strict::String
    # Vorname
    attribute :vorname, Types::Strict::String
    # Strasse mit Nummer
    attribute :strasse, Types::Strict::String.optional.default(nil)
    # Zusatzadresse
    attribute :zusatz, Types::Strict::String.optional.default(nil)
    # Land (CH / D / F / ...)
    attribute :land, Types::Strict::String
    # Postleitzahl
    attribute :plz, Types::Strict::String
    # Ort
    attribute :ort, Types::Strict::String
    # Geburtsdatum
    attribute :geburtsdatum, Types::Strict::Date
  end

  class InvoiceItem < Struct
    Art = Types::Integer.enum(
      0 => 'leistung',
      1 => 'medikament',
      2 => 'labo',
      3 => 'diverse',
    )

    # Datum der Leistung
    attribute :datum, Types::Strict::Date
    # Anzahl der Leistung
    attribute :anzahl, Types::Strict::Integer
    # Zahnnummer der Leistung, Freie Eingabe des Zahnarztes.
    attribute :zahnnummer, Types::Strict::String.optional.default(nil)
    # Art der Leistung (0=Leistung, 1=Medikament, 2=Labo, 3=Diverse)
    attribute :art, Art
    # Nummer der Leistung
    attribute :nummer, Types::Strict::String.optional.default(nil)
    # Bezeichnung in der Sprache der Rechnung
    attribute :bezeichnung, Types::Strict::String
    # Ansatz der Leistung
    attribute :ansatz, Types::Strict::Float.optional.default(nil)
    # Punkte der Leistung
    attribute :punkte, Types::Strict::Float.optional.default(nil)
    # Auf der Leistung ist ein Rabatt gewährt (0 = ohne Rabatt; 1 = mit Rabatt)
    attribute :rabatt, Types::Strict::Bool.default(false)
    # Betrag
    attribute :betrag, Types::Strict::Float
    # Total
    attribute :total, Types::Strict::Float
  end

  class Invoice < Struct
    BehandlungStatus = Types::Integer.enum(
      0 => 'laufend',
      1 => 'abgeschlossen',
    )
    BehandlungTyp = Types::String.enum(
      'PP',
      'VS',
      'IV',
      'KK',
    )
    Sprache = Types::Integer.enum(
      1 => 'DE',
      2 => 'FR',
      3 => 'IT',
      4 => 'EN',
    )

    # Rechnungsnummer aus dem Programm
    attribute :nummer, Types::Strict::Integer
    # Datum der Rechnung (kann auch vordatiert sein)
    attribute :datum, Types::Strict::Date
    # Behandlungsbeginn
    attribute :behandlungsbeginn, Types::Strict::Date
    # Behandlungsende
    attribute :behandlungsende, Types::Strict::Date
    # 0=Behandlung laufend, 1=Behandlung abgeschlossen
    attribute :behandlungstatus, BehandlungStatus
    # Typ der Behandlung (PP / VS / IV / KK)
    attribute :behandlungstyp, BehandlungTyp
    # Sprache der Rechnung (1=deu, 2=franz. 3=ital. 4=engl.)
    attribute :sprache, Sprache
    # Rechnungsüberschrift
    attribute :rech_titel, Types::Strict::String.optional.default(nil)
    # Rechnungstext pro Zeile ein Eintrag
    attribute :rech_text, Types::Strict::String.optional.default(nil)
    # Versicherungsnummer
    attribute :versicherungsnummer, Types::Strict::String.optional.default(nil)
    # Rabatt in Prozent auf den markierten Leistungen
    attribute :rabatt, Types::Strict::Integer.optional.default(nil)
    # Datum der Verfügung
    attribute :verfugungs_datum, Types::Strict::Date.optional.default(nil)
    # Nummer der Verfügung
    attribute :verfugungs_nummer, Types::Strict::Integer.optional.default(nil)
    # Anhang
    attribute :anhange, Types::Strict::Array.of(Types::Strict::String).default([].freeze)

    # Rechnungsbetrag (ohne Vorauszahlungen)
    attribute :total_rabatt, Types::Strict::Float
    # Total aller Vorauszahlungen dieser Rechnung
    attribute :total_vz, Types::Strict::Float
    # Total von allen Leistungen
    attribute :total_leis, Types::Strict::Float
    # Total von allen Medikamenten
    attribute :total_med, Types::Strict::Float
    # Total von allen Labors
    attribute :total_labor, Types::Strict::Float
    # Total diverse Positionen
    attribute :total_diverse, Types::Strict::Float
    # Total minus Vorauszahlungen dieser Rechnung
    attribute :total_netto, Types::Strict::Float
    # # Rechnungsbetrag (ohne Vorauszahlungen)
    attribute :rechnungsbetrag, Types::Strict::Float

    attribute :empfanger, Recipient.optional.default(nil)
    attribute :patient, Patient
    attribute :leistungen, Types::Strict::Array.of(InvoiceItem)

    def total_punkte
      leistungen.select{|l| l.punkte != nil }.map(&:punkte).reduce(0.0, :+)
    end

    def to_h
      h = {
        RechNr: nummer,
        RechDat: datum.strftime('%d.%m.%Y'),
        BehBeginn: behandlungsbeginn.strftime('%d.%m.%Y'),
        BehEnde: behandlungsende.strftime('%d.%m.%Y'),
        BehStatus: behandlungstatus,
        BehTyp: behandlungstyp,
        RechBetrag: rechnungsbetrag,
        TotalLeis: total_leis,
        TotalMed: total_med,
        TotalLabor: total_labor,
        TotalDiverse: total_diverse,
        TotalRabatt: total_rabatt,
        TotalVZ: total_vz,
        TotalNetto: total_netto,
        Sprache: sprache,
        PatNr: patient.nummer,
        PatAnrede: fmt_string(patient.anrede, max: 4),
        PatTitel: fmt_string(patient.titel, max: 20),
        PatName: fmt_string(patient.name, max: 40),
        PatVorname: fmt_string(patient.vorname, max: 40),
        PatStrasse: fmt_string(patient.strasse, max: 30),
        PatZusatz: fmt_string(patient.zusatz, max: 20),
        PatLand: fmt_string(patient.land, max: 3),
        PatPLZ: fmt_string(patient.plz, max: 10),
        PatOrt: fmt_string(patient.ort, max: 40),
        PatGebDatum: fmt_date(patient.geburtsdatum),
      }
      h[:RechTitel] = rech_titel[0..180] unless rech_titel.nil?
      h[:VersNr] = versicherungsnummer unless versicherungsnummer.nil?
      h[:Rabatt] = rabatt unless rabatt.nil?
      h[:VerfügungsDat] = verfugungs_datum unless verfugungs_datum.nil?
      h[:VerfügungsNr] = verfugungs_nummer unless verfugungs_nummer.nil?
      if empfanger
        h[:EmpfNr] = empfanger.nummer
        h[:EmpfArt] = fmt_string(empfanger.art, max: 20)
        h[:EmpfAnrede] = fmt_string(empfanger.anrede, max: 4)
        h[:EmpfTitel] = fmt_string(empfanger.titel, max: 20)
        h[:EmpfName] = fmt_string(empfanger.name, max: 40)
        h[:EmpfVorname] = fmt_string(empfanger.vorname, max: 40)
        h[:EmpfStrasse] = fmt_string(empfanger.strasse, max: 20)
        h[:EmpfLand] = fmt_string(empfanger.land, max: 3)
        h[:EmpfPLZ] = fmt_string(empfanger.plz, max: 10)
        h[:EmpfOrt] = fmt_string(empfanger.ort, max: 40)
        h[:EmpfGebDatum] = fmt_date(empfanger.geburtsdatum)
      end
      anhange.each.with_index(1) do |a, i|
        # Path to attachment file
        h["Anhang#{i}"] = a
      end
      rech_text.scan(/.{1,180}/).each.with_index(1) do |s, i|
        # A line will go up to 180 characters, but could be shorter in case of a
        # new line or simply a short line.
        h["RechText[#{i}]"] = s
      end unless rech_text.nil?
      leistungen.each.with_index(1) do |l, i|
        # Total der Punkt von Ansatz
        # h["TotalPkt[#{i}]"] = l.punkte || 0.0
        # Ansatz für Punkte
        # h["TotalAnsatz[#{i}]"] = l.ansatz || 0.0

        h["LeisDatum[#{i}]"] = fmt_date(l.datum)
        h["LeisAnz[#{i}]"] = fmt_amount(l.anzahl)
        h["LeisZahnNr[#{i}]"] = l.zahnnummer
        h["LeisArt[#{i}]"] = l.art
        h["LeisPosNr[#{i}]"] = l.nummer
        h["LeisBez[#{i}]"] = l.bezeichnung
        h["LeisAnsatz[#{i}]"] = l.ansatz
        h["LeisPkt[#{i}]"] = l.punkte
        h["LeisRabatt[#{i}]"] = l.rabatt ? 1 : 0
        h["LeisBetrag[#{i}]"] = l.betrag
        h["LeisTotal[#{i}]"] = l.total
      end
      h
    end
  end

  # Folgende Einträge sind 1x pro Datei vorhanden
  class Header < Struct
    # Programmname
    attribute :programm, Types::Strict::String
    # Programm Version
    attribute :programm_version, Types::Strict::String
    # Schnittstellenversion
    attribute :version, Types::Strict::String
    # Eindeutige Kundennummer bei Mediserv
    attribute :kundennummer, Types::Strict::Integer
    # Die Stapelnummer ist eine fortlaufende Nummer. Für jede Datei eine neue Nummer
    attribute :stapelnummer, Types::Strict::Integer

    def to_h
      {
        Programm: programm,
        ProgrammVersion: programm_version,
        Version: version,
        KuNr: kundennummer,
        StapelNr: stapelnummer,
      }
    end
  end

  class Batch < Struct
    attribute :allgemein, MediServ::API::Header
    attribute :rechnungen, Types::Strict::Array.of(MediServ::API::Invoice)
    attribute :created_at, Types::Strict::DateTime

    # Anzahl Rechnungen
    def anzahl_rechnungen
      rechnungen.length
    end

    # Gesamtsumme Taxpunkte
    def total_pkt
      rechnungen.map(&:total_punkte).reduce(0.0, :+)
    end

    # Gesamtsumme aller Leistungen
    def total_leis
      rechnungen.map(&:total_leis).reduce(0.0, :+)
    end

    # Gesamtsumme aller Medikamente
    def total_med
      rechnungen.map(&:total_med).reduce(0.0, :+)
    end

    # Gesamtsumme aller Laborkosten
    def total_labor
      rechnungen.map(&:total_labor).reduce(0.0, :+)
    end

    # Gesamtsumme Diverse Positionen
    def total_diverse
      rechnungen.map(&:total_diverse).reduce(0.0, :+)
    end

    # Gesamtsumme Rabatte
    def total_rabatt
      rechnungen.map(&:total_rabatt).reduce(0.0, :+)
    end

    # Gesamtsumme der abgezogenen Vorauszahlungen
    def total_vz
      rechnungen.map(&:total_vz).reduce(0.0, :+)
    end

    # Gesamtsumme Total (ist auch die Summe von TotalLeis + TotalMed + TotalLabor + TotalDiverse + TotalRabatt)
    def total
      total_leis + total_med + total_labor + total_diverse + total_rabatt
    end

    # Nettosumme (Total – Vorauszahlung)
    def total_netto
      total - total_vz
    end

    def to_h
      h = {
        Allgemein: allgemein.to_h.merge({
          Total: total,
          TotalPkt: total_pkt,
          TotalLeis: total_leis,
          TotalMed: total_med,
          TotalLabor: total_labor,
          TotalDiverse: total_diverse,
          TotalRabatt: total_rabatt,
          TotalVz: total_vz,
          TotalNetto: total_netto,
          Datum: fmt_date(created_at),
          Zeit: fmt_time(created_at),
          AnzRech: anzahl_rechnungen,
        }),
      }
      rechnungen.each.with_index(1) do |r,i|
        h["R#{i}"] = r.to_h
      end
      stringify_keys h
    end

    def stringify_keys(h)
      h.reduce({}) do |h, (k,v)|
        v = stringify_keys(v) if v.kind_of?(Hash)
        h.merge({k.to_s => v})
      end
    end
  end

end
